"""Responsible for coloring the text (ANSI stuff)"""

from typing import List

# ANSI Attribute codes
ATTRIBUTES = {
    "bold": 1,
    "dark": 2,
    "italic": 3,
    "underline": 4,
    "blink": 5,
    "reverse": 7,
    "concealed": 8,
    "strike": 9,
    "double-underline": 21,
    "overline": 53,
}

# Standard Foreground Colors (4-bit)
COLORS = {
    "black": 30,
    "grey": 30,  # Actually black but kept for backwards compatibility
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "light_grey": 37,
    "dark_grey": 90,
    "light_red": 91,
    "light_green": 92,
    "light_yellow": 93,
    "light_blue": 94,
    "light_magenta": 95,
    "light_cyan": 96,
    "white": 97,
}


def get_color(color: str, mode: str = "fg") -> int:
    """Returns color code of `color` based on mode `fg` or `bg`"""
    assert mode in ["fg", "bg"], f"mode '{mode}' is not a valid mode."

    if mode == "fg":
        return COLORS[color]
    return COLORS[color] + 10


def colored(
    text: str,
    fgcolor: str = None,
    bgcolor: str = None,
    attrs: List[str] = None,
) -> str:
    """Colorize and return text."""
    assert isinstance(text, str), "text must be str"
    assert text, "text must not be empty"

    # Ansi Format string
    FMT_STR = "\033[%dm%s"

    result = text

    # Apply foreground color
    if fgcolor:
        result = FMT_STR % (get_color(fgcolor), result)

    # Apply background color
    if bgcolor:
        result = FMT_STR % (get_color(bgcolor, "bg"), result)

    # Apply attributes
    if attrs:
        for attr in attrs:
            result = FMT_STR % (ATTRIBUTES[attr], result)

    return result + "\033[0m"


def printc(
    text: str,
    fgcolor: str = None,
    bgcolor: str = None,
    attrs: List[str] = None,
):
    """Colorize and then print text"""
    print(colored(text, fgcolor, bgcolor, attrs))



"""Tests for maketree/core/normalizer.py"""

from os.path import normpath
from maketree.core.normalizer import Normalizer
from maketree.core.parser import Parser


def test_normalize():
    src = """
src/
    file.txt
    file.json
README.md
"""
    parsed_tree = Parser._parse_lines(src.splitlines())
    expected_paths = {
        "directories": [
            str(normpath("root/src/")),
        ],
        "files": [
            str(normpath("root/src/file.txt")),
            str(normpath("root/src/file.json")),
            str(normpath("root/README.md")),
        ],
    }

    assert Normalizer.normalize(parsed_tree, "root") == expected_paths



"""Frontend of the project (Argument handling and stuff)"""

import sys
from pathlib import Path
from argparse import ArgumentParser
from maketree.core.parser import Parser, ParseError
from maketree.core.tree_builder import TreeBuilder
from maketree.core.normalizer import Normalizer
from maketree.terminal_colors import colored
from maketree.console import Console
from maketree.utils import (
    is_valid_dirpath,
    get_existing_paths,
    print_tree,
    create_dir,
)
from typing import List, Dict, Tuple


PROGRAM = "maketree"
VERSION = "1.1.0"


def main():
    args = parse_args()

    sourcefile = Path(args.src)
    dstpath = Path(args.dst)
    CREATE_DST = args.create_dst
    VERBOSE: bool = args.verbose
    OVERWRITE: bool = args.overwrite
    SKIP: bool = args.skip
    PRINT_TREE = args.graphical
    NO_COLORS = args.no_color
    NO_CONFIRM = args.no_confirm

    # Console? (is this fuc**ing Yavascript?)
    console = Console(VERBOSE, NO_COLORS)

    # Mutually Exclusive
    if OVERWRITE and SKIP:
        console.error(
            console.color_substrs(
                "Options --overwrite and --skip are mutually exlusive. ",
                ["--overwrite", "--skip"],
                "light_yellow",
            )
        )

    # SRC Exists?
    if not sourcefile.exists():
        console.error("source '%s' does not exist." % sourcefile)

    # SRC Tree file?
    if not sourcefile.name.endswith(".tree"):
        console.error("source '%s' is not a .tree file." % sourcefile)

    # DST Exists?
    if not dstpath.is_dir():
        if CREATE_DST:
            console.verbose("Validating '%s'..." % dstpath)
            valid = is_valid_dirpath(dstpath)
            if valid is not True:
                console.error(valid)
        else:
            console.error(
                console.color_substrs(
                    "destination path '%s' does not exist." % dstpath,
                    [dstpath],
                    "light_red",
                )
            )

    # Parse the source file
    console.verbose("Parsing %s..." % sourcefile)
    try:
        parsed_tree = Parser.parse_file(sourcefile)
    except ParseError as e:
        console.error(e)

    # Print the graphical tree and Exit.
    if PRINT_TREE:
        print_tree(parsed_tree, root=dstpath, console=console)
        sys.exit(0)

    # Confirm before proceeding
    if not NO_CONFIRM:
        print_tree(parsed_tree, root=dstpath, console=console)
        proceed: bool = console.input_confirm(
            "Create this structure? (y/N): ", fgcolor="light_magenta"
        )
        if not proceed:
            sys.exit(0)

    console.verbose("Creating tree paths...")

    # Create paths from tree nodes
    paths: Dict[str, List[str]] = Normalizer.normalize(parsed_tree, dstpath)

    # If Overwrite and Skip both are false
    if not OVERWRITE and not SKIP:
        # Check existing paths
        console.verbose("Checking existing paths...\n")
        existing_paths = get_existing_paths(paths["files"])
        count = len(existing_paths)
        # Any path exists?
        if count:
            console.print_lines(
                existing_paths,
                "Warning: File already exists: ",
                color="light_yellow",
                force_print=False,
            )
            print()
            console.error(
                console.color_substrs(
                    f"Found {count} existing files, cannot proceed. "
                    "(try --skip or --overwrite)",
                    ["--skip", "--overwrite"],
                    "light_yellow",
                )
            )

    # Create dstpath here...
    console.verbose("Creating '%s'..." % dstpath)
    created = create_dir(dstpath)
    if created is not True:
        console.error(created)

    console.verbose("Creating tree in '%s'...\n" % dstpath)

    # Create the files and dirs finally
    build_count = TreeBuilder.build(
        paths,
        console,
        skip=SKIP,
        overwrite=OVERWRITE,
    )

    # Completion message
    built_dirs = f"{build_count[0]} directories"
    built_files = f"{build_count[1]} files"

    print(
        console.color_substrs(
            f"\n{built_dirs} and {built_files} have been created.",
            [built_dirs, built_files],
            "light_green",
        )
    )


def parse_args():
    """Parse command-line arguments and return."""

    parser = ArgumentParser(
        prog=PROGRAM,
        usage="%(prog)s [OPTIONS]",
        epilog="%s %s" % (PROGRAM.title(), VERSION),
        description="Create complex project structures effortlessly.",
    )

    parser.add_argument("src", help="source file (with .tree extension)")
    parser.add_argument(
        "dst",
        nargs="?",
        default=".",
        help="where to create the tree structure (default: %(default)s)",
    )
    parser.add_argument(
        "-cd",
        "--create-dst",
        action="store_true",
        help="create destination folder if it doesn't exist.",
    )
    parser.add_argument(
        "-g",
        "--graphical",
        action="store_true",
        help="show source file as graphical tree and exit",
    )
    parser.add_argument(
        "-o", "--overwrite", action="store_true", help="overwrite existing files"
    )
    parser.add_argument("-s", "--skip", action="store_true", help="skip existing files")
    parser.add_argument(
        "-nc",
        "--no-color",
        action="store_true",
        help="don't use colors in output",
    )
    parser.add_argument(
        "-nC",
        "--no-confirm",
        action="store_true",
        help="don't ask for confirmation",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="enable verbose mode"
    )

    return parser.parse_args()



"""Tests for maketree/console.py"""

from maketree.console import Console

console = Console(verbose=False, no_color=True)


def test_colored():
    """This Function only colors a string when NO_COLOR is False"""

    s = "This is a dummy string."
    assert console.colored(s, fgcolor="light_yellow", bgcolor="dark_grey") == s

    console.NO_COLOR = False
    assert console.colored(s, fgcolor="light_yellow", bgcolor="dark_grey") != s


def test_color_substrs():
    """This function colors all sub-strings found in a string."""
    # Doesn't Color, When NO_COLOR set to True
    console.NO_COLOR = True
    s = "This is a dummy string."
    colored_s = console.color_substrs(s, substrs=["is", "in"], fgcolor="green")
    assert colored_s == s

    # Colors, When NO_COLOR set to False
    console.NO_COLOR = False
    colored_s = console.color_substrs(s, substrs=["is", "in"], fgcolor="green")
    assert (
        colored_s
        == "Th\x1b[32mis\x1b[0m \x1b[32mis\x1b[0m a dummy str\x1b[32min\x1b[0mg."
    )



"""Contains logic for creating the directory structure on the file system,
based on the parsed data from the structure file."""

import os
from typing import List, Dict, Tuple, Optional
from maketree.console import Console


class TreeBuilder:
    """Build the tree parsed from `.tree` file"""

    @classmethod
    def build(
        cls,
        paths: Dict[str, List[str]],
        console: Optional[Console] = None,
        skip: bool = False,
        overwrite: bool = False,
    ) -> Tuple[int, int]:
        """
        ### Build
        Create the directories and files on the filesystem.

        #### Args:
        - `paths`: the paths dictionary
        - `skip`: skips existing files
        - `overwrite`: overwrites existing files
        - `verbose`: print messages while creating dirs/files
        - `no_color`: print messages without colors

        Returns a `tuple[int, int]` containing the number of
        dirs and files created, in that order.
        """
        # Console instance from CLI
        cls.console = console

        # Create directories
        dirs_created = cls.create_dirs(paths["directories"])

        # Create Files
        files_created = cls.create_files(
            paths["files"],
            skip=skip,
            overwrite=overwrite,
        )

        return (dirs_created, files_created)

    @classmethod
    def create_dirs(cls, dirs: List[str]) -> int:
        """Create files with names found in `files`.
        Returns the number of dirs created."""
        count = 0
        for path in dirs:
            try:
                os.mkdir(path)  # Create the directory
                count += 1
                cls.console.print("[D] Creating '%s'" % path, "light_green")

            except FileExistsError:
                cls.console.print(
                    "[D] Skipping '%s', already exists" % path,
                    "light_yellow",
                )
        return count

    @classmethod
    def create_files(
        cls,
        files: List[str],
        skip: bool = False,
        overwrite: bool = False,
    ) -> int:
        """Create files with names found in `files`. Returns the number of files created."""
        count = 0
        for path in files:
            try:
                # Create file
                with open(path, "x") as _:
                    cls.console.print("[f] Creating '%s'" % path, "light_green")

                count += 1
            except FileExistsError:
                # Skip file
                if skip:
                    cls.console.print(
                        "[F] Skipping '%s', already exists" % path,
                        "light_yellow",
                    )
                    continue

                # Overwrite file
                if overwrite:
                    count += 1
                    cls.console.print("[F] Overwriting '%s'" % path, "light_blue")
                    with open(path, "w") as _:
                        continue

        return count



"""Tests for maketree/utils.py"""

import os
import shutil

from maketree.utils import (
    is_valid_dirpath,
    _contains,
    contains_chars,
    is_valid_dir,
    is_valid_file,
    is_valid_extension,
    get_existing_paths,
    get_nonexisting_paths,
)


# Create temporary files/folders inside this and delete aftwards
TEMP_DIR = "temp"

paths = [
    os.path.join(TEMP_DIR, "file.txt"),
    os.path.join(TEMP_DIR, "file.json"),
    os.path.join(TEMP_DIR, ".gitignore"),
    os.path.join(TEMP_DIR, "LICENSE"),
]


def test_get_existing_paths():
    os.mkdir(TEMP_DIR)

    # Create first-two files in paths list.
    for p in paths[:2]:
        with open(p, "x") as _:
            pass

    # Get nonexisting paths...
    try:
        get_nonexisting_paths(paths) == paths[2:]
    finally:
        # Remove TEMP_DIR
        shutil.rmtree(TEMP_DIR)


def test_get_existing_paths():
    try:
        os.mkdir(TEMP_DIR)
    except FileExistsError:
        pass

    # Create First-two files in paths list.
    for p in paths[:2]:
        with open(p, "x") as _:
            pass

    # Get existings paths now...
    try:
        assert get_existing_paths(paths) == paths[:2]
    finally:
        # Remove TEMP_DIR
        shutil.rmtree(TEMP_DIR)


def test_is_valid_extension():
    assert is_valid_extension(".txt") == True
    assert is_valid_extension(".gitignore") == True
    assert is_valid_extension(".c") == True
    assert is_valid_extension(".🔥") == True

    assert is_valid_extension("file.txt") == False  # Accepts only extensions
    assert is_valid_extension("...txt") == False  # Too many periods
    assert is_valid_extension(".t/xt") == False  # slash
    assert is_valid_extension("txt") == False  # must start with period


def test_is_valid_file():
    assert is_valid_file("file.extension") == True
    assert is_valid_file("filewithoutextension") == True
    assert is_valid_file(".gitignore") == True
    assert is_valid_file("file 123.txt") == True
    assert is_valid_file(".") == True

    assert isinstance(is_valid_file(""), str)
    assert isinstance(is_valid_file("fi:le.txt"), str)

    # Only single part names are valid
    assert isinstance(is_valid_file("fi/le.txt"), str)


def test_is_valid_dir():
    assert is_valid_dir("folder") == True
    assert is_valid_dir("folder123") == True
    assert is_valid_dir("folder 134") == True
    assert is_valid_dir(".folder") == True
    assert is_valid_dir(".") == True

    assert isinstance(is_valid_dir("fol/der"), str)  # Only single part names are valid
    assert isinstance(is_valid_dir("fol:der"), str)


def test_contains_chars():
    assert contains_chars("ABCD", "C") == True
    assert contains_chars("ABCD", "c") == False

    assert contains_chars("1.2.1.3", ".,") == True  # Either . or ,
    assert contains_chars("1.2.1.3", ",:") == False  # Either , or :

    assert contains_chars("example@gmail.com", "(@)") == True  # Any of ( @ )
    assert contains_chars("example@gmail.com", "(:)") == False  # Any of ( : )

    assert contains_chars("abc def", " ") == True
    assert contains_chars("abcdef", " ") == False


def test_is_valid_dirpath():
    assert is_valid_dirpath(".") == True
    assert is_valid_dirpath("folder") == True
    assert is_valid_dirpath("./folder/") == True
    assert is_valid_dirpath("folder/folder1/folder2/folder3/folder/4") == True
    assert is_valid_dirpath("./folder/") == True

    assert isinstance(is_valid_dirpath("./fol<der>/"), str)
    assert isinstance(is_valid_dirpath("./fol<der/fo>lder/"), str)
    assert isinstance(is_valid_dirpath(""), str)


def test_contains():
    parts = [
        "abcd",
        "abcde",
        "testing",
        "testing123",
    ]

    assert _contains(parts, "2td") == True
    assert _contains(parts, "2td") == True
    assert _contains(parts, "_:") == False  # None of parts contain _ or :
    assert _contains(parts, "xz") == False  # None of parts contain x or z



"""Tests for maketree/terminal_colors.py"""

from pytest import raises
from maketree.terminal_colors import colored, get_color


def test_get_color():
    assert get_color("green", "fg") == 32
    assert get_color("yellow", "fg") == 33
    assert get_color("light_cyan", "fg") == 96
    assert get_color("white", "fg") == 97

    # All background colors 10 ahead of their foregrounds
    assert get_color("green", "bg") == 32 + 10
    assert get_color("yellow", "bg") == 33 + 10
    assert get_color("light_cyan", "bg") == 96 + 10
    assert get_color("white", "bg") == 97 + 10


def test_colored():
    # No matter what, reset code gets added (takes `if` to avoid, unnecessary)

    assert colored("This is string.", "blue") == "\x1b[34mThis is string.\x1b[0m"
    assert (
        colored("This is string.", "blue", "dark_grey")
        == "\x1b[100m\x1b[34mThis is string.\x1b[0m"
    )

    with raises(AssertionError):
        assert colored(4, "blue") == "\033[0m"

    with raises(AssertionError):
        assert colored("") == "\033[0m"



"""Contains Helper code to keep core logic clean. (things that don't fit anywhere, fit here)"""

from os import makedirs
from os.path import exists, splitext
from pathlib import Path
from platform import system
from typing import List, Dict, Union, Iterable, Optional
from maketree.terminal_colors import colored
from maketree.console import Console


# Windows, Darwin, Linux
OS: str = system()


def get_nonexisting_paths(paths: List[str]) -> List[str]:
    """Returns a list of non-existing paths from `paths` list."""
    return list(filter(lambda p: not exists(p), paths))


def get_existing_paths(paths: List[str]) -> List[str]:
    """Returns a list of existing paths from `paths` list."""
    return list(filter(lambda p: exists(p), paths))


def is_valid_extension(extension: str) -> bool:
    """
    ### Is Valid Extension
    Returns `True` if extension is valid, `False` otherwise.
    `extension` must contain a period `.`

    An extension is valid if it follows below criteria:
    - extension must be non-empty (excluding period `.`)
    - extension must have a period `.`
    - extension must not contain symbols, whitespaces
        - `\\/:*?"<>|` are illegal on Windows
        - `/:` are illegal on Mac
        - `/` are illegal on Linux
    """
    if not extension:
        return False

    if len(extension) < 2:
        return False

    if not extension.startswith("."):
        return False

    if extension.count(".") > 1:
        return False

    if OS == "Windows":
        # Got Illegals?
        if contains_chars(extension, r' \/:*?"<>|'):
            return False

    elif OS == "Darwin" and contains_chars(extension, "/:"):
        return False

    elif OS == "Linux" and "/" in extension:
        return False

    return True


def is_valid_file(filename: str) -> Union[bool, str]:
    """
    ### Is Valid File
    Validates filename. Returns `True` if valid, Returns `str` if invalid.
    This `str` is the cause of filename invalidation.

    #### ARGS:
    - `filename`: name of the file

    #### Note:
    This function is not a stripped down version of itself. (specific to needs of the `Parser`, minimal but fast)
    """
    if not filename:
        return "file name must not be empty"

    # Split filepath into root and extension
    root, ext_ = splitext(filename)

    # Root must not be empty
    if not root:
        return "invalid file name"

    # Check for illegal chars
    if OS == "Windows":
        if contains_chars(root, r'\/:?*<>"|'):
            return 'avoid these illegal characters: \\/:?*<>|"'
    elif OS == "Darwin":
        if contains_chars(root, r"/:<>"):
            return "avoid these illegal characters: /:?<>"
    else:  # Linux
        if contains_chars(root, r"/:<>"):
            return "avoid these illegal characters: /:?<>"

    if ext_ and not is_valid_extension(ext_):
        return "invalid file extension"

    return True


def is_valid_dir(dirname: str) -> Union[bool, str]:
    """
    ### Is Valid Dirpath
    Validates directory name. Returns `True` if valid, Returns `str` if invalid.
    This `str` contains the reason for dir being invalid.

    #### ARGS:
    - `dirname`: the path to validate

    #### Note:
    This function is not a stripped down version of itself. (specific to needs of the `Parser`, minimal but fast)
    """
    if not dirname:
        return "path must not be empty."

    # Check for illegal chars
    if OS == "Windows":
        if contains_chars(dirname, r'\/:?*<>"|'):
            return 'avoid these illegal characters: \\/:?*<>|"'
    elif OS == "Darwin":
        if contains_chars(dirname, r"/:<>"):
            return "avoid these illegal characters: /:?<>"
    else:
        if contains_chars(dirname, r"/:<>"):
            return "avoid these illegal characters: /:?<>"

    return True


def is_valid_dirpath(dirpath: str, max_length: int = 250):
    """
    ### Is Valid Dirpath
    Validates directory path. Returns `True` if valid, Returns `str` if invalid.
    This `str` contains the reason for path being invalid.

    #### ARGS:
    - `dirpath`: the path to validate
    - `max_length`: maximum length to allow (length of the whole path, except drive)

    #### Example:
    ```
    >> is_valid_dirpath("path\\to\\folder")
    True
    >> is_valid_dirpath("path\\to\\*Illegal*folder")
    'Illegal characters are not allowed: \\/:?*<>|"'
    ```

    Raises `AssertionError` if:
    - `dirpath` is not a string

    Used for longer dir paths.
    """
    if not dirpath:
        return "path must not be empty."

    d = Path(dirpath)
    if d.drive:
        root_parts = d.parts[1:]
    elif OS == "Linux" and (d.parts and d.parts[0] == "/"):
        root_parts = d.parts[1:]
    else:
        root_parts = d.parts

    if sum(len(part) for part in root_parts) > max_length:
        return (
            f"maximum length of path can be {max_length} (excluding slashes and drive)"
        )

    # Check for illegal chars
    if OS == "Windows":
        if _contains(root_parts, r'\/:?*<>"|'):
            return 'avoid these characters: \\/:?*<>|"'
    elif OS == "Darwin":
        if _contains(root_parts, r"/:<>"):
            return "avoid these characters: /:<>"
    else:
        if _contains(root_parts, r"/:<>"):
            return "avoid these characters: /:<>"

    return True


def _contains(parts: Iterable[str], chars: str) -> bool:
    """
    ### Contains
    Checks whether a string in `parts` contains a character from `chars`.
    Returns `True` if it does, `False` if does not.

    Used with `is_valid_dirpath` only.
    """
    for char in chars:
        for part in parts:
            if char in part:
                return True
    return False


def contains_chars(string: str, chars: str) -> bool:
    """
    ### Contains
    Checks whether `string` contains a character from `chars`.
    Returns `True` if it does, `False` if does not.

    Used with `is_valid_dir`.
    """
    return any(char for char in chars if char in string)


def print_tree(tree: List[Dict], console: Console, root: str = "."):
    """Prints the parsed `tree` in a graphical format. _(Not perfect but, gets the job done)_"""
    tab = 0
    BAR = console.colored("│   ", "dark_grey")
    LINK = console.colored("├───", "dark_grey")
    LINK_LAST = console.colored("└───", "dark_grey")
    FMT_STR = f"%s%s %s"

    def traverse(node: Dict, childs: int):
        nonlocal tab
        count = 0  # keeps track of child counts

        for child in node.get("children", []):
            count += 1

            child_name = child["name"]

            # Add a Slash '/' after a directory
            if child["type"] == "directory":
                child_name = console.colored(
                    "%s/" % child_name,
                    fgcolor="light_green",
                    attrs=["italic", "bold"],
                )

            if count == childs:
                # Last Child
                print(FMT_STR % (BAR * tab, LINK_LAST, child_name))
            else:
                # Others
                print(FMT_STR % (BAR * tab, LINK, child_name))

            if child["type"] == "directory" and child["children"]:
                tab += 1
                traverse(child, len(child["children"]))
        tab -= 1
        return

    root = str(root) if str(root) == "." else f"{root}/"
    print(
        console.colored(
            root,
            fgcolor="light_green",
            attrs=["italic", "bold"],
        )
    )

    traverse(
        node={
            "type": "directory",
            "name": root,
            "children": tree,
        },
        childs=len(tree),
    )


def create_dir(path: str):
    """Create a folder/directory on the filesystem."""
    # Create folder
    try:
        makedirs(path, exist_ok=True)
        return True
    except Exception as e:
        return str(e)



"""Tests for maketree/core/parser.py"""

from maketree.core.parser import Parser


def test_parse_lines():
    structure = """
src/
    file1.txt
    file2.txt
LICENSE
README.md
"""
    expected_tree = [
        {
            "type": "directory",
            "name": "src",
            "children": [
                {"type": "file", "name": "file1.txt"},
                {"type": "file", "name": "file2.txt"},
            ],
        },
        {"type": "file", "name": "LICENSE"},
        {"type": "file", "name": "README.md"},
    ]

    assert Parser._parse_lines(structure.splitlines()) == expected_tree
    assert Parser._parse_lines(["", "", ""]) == []  # Empty lines



"""Tests for maketree/core/tree_builder.py"""

import shutil
from os.path import exists
from os import mkdir
from maketree.core.parser import Parser
from maketree.core.normalizer import Normalizer
from maketree.core.tree_builder import TreeBuilder
from maketree.console import Console

TEMP_DIR = "temp"


def test_build():
    console = Console(False, True)
    src = """ 
src/
    file.txt
    file.py
README.md
.gitignore
"""
    parsed_tree = Parser._parse_lines(src.splitlines())
    paths = Normalizer.normalize(parsed_tree, rootpath=TEMP_DIR)

    # Make temp directory
    try:
        mkdir(TEMP_DIR)
    except FileExistsError:
        pass

    # Build the tree
    build_count = TreeBuilder.build(paths, console=console)

    assert build_count[0] == len(paths["directories"])
    assert build_count[1] == len(paths["files"])

    # Check if they are actually created.
    assert exists(paths["directories"][0]) == True
    assert exists(paths["files"][0]) == True
    assert exists(paths["files"][1]) == True
    assert exists(paths["files"][2]) == True
    assert exists(paths["files"][3]) == True

    # Remove temp directory
    shutil.rmtree(TEMP_DIR)



import sys
from maketree.terminal_colors import printc, colored
from typing import List, Optional


class Console:
    """
    ### Console
    class for Input/Output

    #### ARGS:
    - `verbose`: decides whether to print verbose messages or not
    - `no_color`: decides whether to use colors in output or not

    """

    def __init__(
        self,
        verbose,
        no_color,
    ):
        self.VERBOSE = verbose
        self.NO_COLOR = no_color

        self.clr_info = "light_blue"
        self.clr_error = "light_red"
        self.clr_success = "light_green"
        self.clr_warning = "light_yellow"
        self.clr_primary = "light_magenta"
        self.clr_secondary = "yellow"

    def error(self, message: str):
        """Print `message` and exit with status `1`. Use for errors only."""
        self.print(
            self.color_substrs(
                "Error: %s" % message,
                ["Error:"],
                self.clr_error,
            ),
            force_print=True,
        )
        sys.exit(1)

    def info(self, message: str):
        """Print `message`. Use for informational messages."""
        self.print(
            self.color_substrs(
                "[INFO] %s" % message,
                ["[INFO]"],
                self.clr_info,
            ),
            force_print=True,
        )

    def verbose(self, message: str):
        """Print `message`. Use for verbose messages."""
        self.print(
            self.color_substrs(
                "[INFO] %s" % message,
                ["[INFO]"],
                self.clr_info,
            ),
        )

    def warning(self, message: str):
        """Print `message`. Use for warning messages."""
        self.print(
            self.color_substrs(
                "Warning: %s" % message,
                ["Warning:"],
                self.clr_warning,
            ),
            force_print=True,
        )

    def success(self, message: str):
        """Print `message`. Use for success messages."""
        self.print(
            self.color_substrs(
                "Success: %s" % message,
                ["Success:"],
                self.clr_success,
            ),
            force_print=True,
        )

    def print(
        self,
        text: str,
        fgcolor: Optional[str] = None,
        bgcolor: Optional[str] = None,
        attrs: Optional[List[str]] = None,
        *,
        force_print: bool = False,
        sep: Optional[str] = " ",
        end: str = "\n",
        flush: bool = False,
    ):
        """
        ### Print
        Custom print function that prints `text`.

        ####  ARGS:
        - `text`: the text to print
        - `fgcolor`: foreground color of text
        - `bgcolor`: background color of text
        - `attrs`: attributes to apply to text
        - `force_print`: overrides VERBOSE, force prints text
        """
        if not force_print and not self.VERBOSE:
            return

        if self.NO_COLOR:
            print(text, sep=sep, end=end, flush=flush)
            return

        print(colored(text, fgcolor, bgcolor, attrs), sep=sep, end=end, flush=flush)

    def print_lines(
        self,
        lines: List[str],
        prefix: str = "",
        suffix: str = "",
        color: Optional[str] = None,
        *,
        force_print: bool = False,
    ):
        """
        Print lines (list of strings).

        #### Args:
        - `lines`: list of strings
        - `prefix`: prefix string to add at the start of each line
        - `suffix`: suffix string to add at the end of each line
        - `color`: foreground color of strings
        - `force_print`: decides whether to print, or leave it to VERBOSE
        """
        for line in lines:
            self.print(
                "%s%s%s" % (prefix, line, suffix),
                fgcolor=color,
                force_print=force_print,
            )

    def input_confirm(self, message: str, fgcolor: Optional[str] = None) -> bool:
        """Confirm and return `true` or `false`"""
        while True:
            try:
                self.print(message, fgcolor=fgcolor, force_print=True, end="")
                answer = input().lower()
                if answer == "y" or answer == "yes":
                    return True
                elif answer == "n" or answer == "no":
                    return False

                # Otherwise just repeat

            except KeyboardInterrupt:
                # Force quit
                sys.exit(1)
            except:
                continue

    def color_substrs(
        self,
        text: str,
        substrs: List[str],
        fgcolor: Optional[str] = None,
    ):
        """
        ### Color Substrings
        Color all sub-strings in `text` and return the colored text.

        #### ARGS:
        - `text`: text to color
        - `substrs`: list of sub-strings
        - `fgcolor`: foreground color of substrings

        """
        if self.NO_COLOR:
            return text

        colored_text = text
        for string in substrs:
            colored_text = colored_text.replace(
                str(string),
                colored(str(string), fgcolor),
            )

        return colored_text

    def colored(
        self,
        text: str,
        fgcolor: str = None,
        bgcolor: str = None,
        attrs: List[str] = None,
    ) -> str:
        """Colorize and return text. Returns text without
        color if `NO_COLOR` is `True`."""
        if self.NO_COLOR:
            return text

        return colored(text, fgcolor, bgcolor, attrs)
